#pragma once

#include <atomic>
#include <cstdint>
#include <memory>
#include <string>
#include <vector>
#include <functional>

#include "prometheus/collectable.h"
#include "prometheus/detail/pull_export.h"
#include "prometheus/registry.h"

class CivetServer;

namespace prometheus {

namespace detail {
class MetricsHandler;
class SerialMetricsHandler;
}  // namespace detail

class PROMETHEUS_CPP_PULL_EXPORT Exposer {
 public:
  explicit Exposer(const std::string& bind_address,
                   const std::string& uri = std::string("/metrics"),
                   const std::size_t num_threads = 2);
  ~Exposer();
  void RegisterCollectable(const std::weak_ptr<Collectable>& collectable);

 private:
  std::unique_ptr<CivetServer> server_;
  std::vector<std::weak_ptr<Collectable>> collectables_;
  std::shared_ptr<Registry> exposer_registry_;
  std::unique_ptr<detail::MetricsHandler> metrics_handler_;
  std::string uri_;
};

class SerialExposer {
public:
	explicit SerialExposer(const std::function<void()> update_callback = std::function<void()> {});
	std::pair<std::map<std::string, std::string>, std::string>
	operator()(const std::map<std::string, std::string>& req_headers);
	void RegisterCollectable(const std::weak_ptr<Collectable>& collectable);
	~SerialExposer();

private:
	std::vector<std::weak_ptr<Collectable>> collectables_;
	std::shared_ptr<Registry> exposer_registry_;
	std::unique_ptr<detail::SerialMetricsHandler> metrics_handler_;
};

}  // namespace prometheus
