#pragma once

#include <memory>
#include <vector>
#include <functional>
#include "CivetServer.h"

#include "prometheus/counter.h"
#include "prometheus/registry.h"
#include "prometheus/summary.h"

namespace prometheus {
namespace detail {
class MetricsHandler : public CivetHandler {
 public:
  MetricsHandler(const std::vector<std::weak_ptr<Collectable>>& collectables,
                 Registry& registry);

  bool handleGet(CivetServer* server, struct mg_connection* conn) override;

 private:
  std::vector<MetricFamily> CollectMetrics() const;

  const std::vector<std::weak_ptr<Collectable>>& collectables_;
  Family<Counter>& bytes_transferred_family_;
  Counter& bytes_transferred_;
  Family<Counter>& num_scrapes_family_;
  Counter& num_scrapes_;
  Family<Summary>& request_latencies_family_;
  Summary& request_latencies_;
  std::function<void()> update_callback_ {};
};

//////////////////////////////////////////////////////
///          Serial Metrics Handler                ///
//////////////////////////////////////////////////////

class SerialMetricsHandler {
 public:
  SerialMetricsHandler(
    const std::vector<std::weak_ptr<Collectable>>& collectables,
    Registry& registry,
    const std::function<void()> update_callback = std::function<void()> {}
  );

  std::pair<std::map<std::string, std::string>, std::string>
  operator()(const std::map<std::string, std::string>& req_headers);

 private:
  std::vector<MetricFamily> CollectMetrics() const;

  const std::vector<std::weak_ptr<Collectable>>& collectables_;
  Family<Counter>& bytes_transferred_family_;
  Counter& bytes_transferred_;
  Family<Counter>& num_scrapes_family_;
  Counter& num_scrapes_;
  Family<Summary>& request_latencies_family_;
  Summary& request_latencies_;
  std::function<void()> update_callback_ {};
};
}  // namespace detail
}  // namespace prometheus
